/*****************************************************************************
                    The Dark Mod GPL Source Code
 
 This file is part of the The Dark Mod Source Code, originally based 
 on the Doom 3 GPL Source Code as published in 2011.
 
 The Dark Mod Source Code is free software: you can redistribute it 
 and/or modify it under the terms of the GNU General Public License as 
 published by the Free Software Foundation, either version 3 of the License, 
 or (at your option) any later version. For details, see LICENSE.TXT.
 
 Project: The Dark Mod (http://www.thedarkmod.com/)
 
 $Revision: 5122 $ (Revision of last commit) 
 $Date: 2011-12-11 21:47:31 +0200 (Sun, 11 Dec 2011) $ (Date of last commit)
 $Author: greebo $ (Author of last commit)
 
******************************************************************************/

#ifndef __BASE64_H__
#define __BASE64_H__

/*
===============================================================================

	base64

===============================================================================
*/

class idBase64 {
public:
				idBase64( void );
				idBase64( const idStr &s );
				~idBase64( void );

	void		Encode( const byte *from, int size );
	void		Encode( const idStr &src );
	int			DecodeLength( void ) const; // minimum size in bytes of destination buffer for decoding
	int			Decode( byte *to ) const; // does not append a \0 - needs a DecodeLength() bytes buffer
	void		Decode( idStr &dest ) const; // decodes the binary content to an idStr (a bit dodgy, \0 and other non-ascii are possible in the decoded content)
	void		Decode( idFile *dest ) const;

	const char	*c_str() const;

	void 		operator=( const idStr &s );

private:
	byte *		data;
	int			len;
	int			alloced;

	void		Init( void );
	void		Release( void );
	void		EnsureAlloced( int size );
};

ID_INLINE idBase64::idBase64( void ) {
	Init();
}

ID_INLINE idBase64::idBase64( const idStr &s ) {
	Init();
	*this = s;
}

ID_INLINE idBase64::~idBase64( void ) {
	Release();
}

ID_INLINE const char *idBase64::c_str( void ) const {
	return (const char *)data;
}

ID_INLINE void idBase64::Init( void ) {
	len = 0;
	alloced = 0;
	data = NULL;
}

ID_INLINE void idBase64::Release( void ) {
	if ( data ) {
		delete[] data;
	}
	Init();
}

ID_INLINE void idBase64::EnsureAlloced( int size ) {
	if ( size > alloced ) {
		Release();
	}
	data = new byte[size];
	alloced = size;
}

ID_INLINE void idBase64::operator=( const idStr &s ) {
	EnsureAlloced( s.Length()+1 ); // trailing \0 - beware, this does a Release
	strcpy( (char *)data, s.c_str() );
	len = s.Length();
}

#endif /* !__BASE64_H__ */
